public inbox for cygwin-cvs@sourceware.org
help / color / mirror / Atom feed
* [newlib-cygwin/main] Cygwin: select: workaround FD_WRITE network event handling
@ 2023-08-01 12:23 Corinna Vinschen
  0 siblings, 0 replies; only message in thread
From: Corinna Vinschen @ 2023-08-01 12:23 UTC (permalink / raw)
  To: cygwin-cvs

https://sourceware.org/git/gitweb.cgi?p=newlib-cygwin.git;h=dedbbd74d0a8f3b7dfae6188321703a47bb8a2b3

commit dedbbd74d0a8f3b7dfae6188321703a47bb8a2b3
Author:     Corinna Vinschen <corinna@vinschen.de>
AuthorDate: Tue Aug 1 14:22:55 2023 +0200
Commit:     Corinna Vinschen <corinna@vinschen.de>
CommitDate: Tue Aug 1 14:22:55 2023 +0200

    Cygwin: select: workaround FD_WRITE network event handling
    
    The FD_WRITE event is a false friend.  It indicates ready to write
    even if the next send fails with WSAEWOULDBLOCK.  *After* the fact,
    FD_WRITE will be cleared until sending is again possible, but that
    is too late for a select/write loop.
    
    Workaround that by using the WinSock select function when peeking
    at a socket and FD_WRITE gets indicated. WinSock select fortunately
    indicates writability correctly.
    
    Fixes: 70e476d27be8 ("(peek_socket): Use event handling for peeking socket.")
    Signed-off-by: Corinna Vinschen <corinna@vinschen.de>

Diff:
---
 winsup/cygwin/autoload.cc |  1 +
 winsup/cygwin/select.cc   | 30 +++++++++++++++++++++++++++---
 2 files changed, 28 insertions(+), 3 deletions(-)

diff --git a/winsup/cygwin/autoload.cc b/winsup/cygwin/autoload.cc
index c9ad92530c83..67149f828266 100644
--- a/winsup/cygwin/autoload.cc
+++ b/winsup/cygwin/autoload.cc
@@ -645,6 +645,7 @@ LoadDLLfunc (getsockname, ws2_32)
 LoadDLLfunc (getsockopt, ws2_32)
 LoadDLLfunc (ioctlsocket, ws2_32)
 LoadDLLfunc (listen, ws2_32)
+LoadDLLfunc_pfx_only (select, ws2_32)
 LoadDLLfunc (setsockopt, ws2_32)
 LoadDLLfunc (shutdown, ws2_32)
 LoadDLLfunc (socket, ws2_32)
diff --git a/winsup/cygwin/select.cc b/winsup/cygwin/select.cc
index bad4c37f3b2e..7b9473849ec2 100644
--- a/winsup/cygwin/select.cc
+++ b/winsup/cygwin/select.cc
@@ -1757,6 +1757,15 @@ fhandler_base::select_except (select_stuff *ss)
   return s;
 }
 
+struct wfd_set
+{
+  u_int   fd_count;
+  SOCKET  fd_array[1];
+};
+
+/* autoload exposes the Winsock select function only with a _win32_ prefix. */
+extern "C" int _win32_select(int, wfd_set *, wfd_set *, wfd_set *, TIMEVAL *);
+
 static int
 peek_socket (select_record *me, bool)
 {
@@ -1772,9 +1781,24 @@ peek_socket (select_record *me, bool)
   if (me->read_selected)
     me->read_ready |= ret || !!(events & (FD_READ | FD_ACCEPT | FD_CLOSE));
   if (me->write_selected)
-    /* Don't check for FD_CLOSE here.  Only an error case (ret == -1)
-       will set ready for writing. */
-    me->write_ready |= ret || !!(events & (FD_WRITE | FD_CONNECT));
+    {
+      /* The FD_WRITE event is a false friend. It indicates ready to write
+         even if the next send fails with WSAEWOULDBLOCK.  *After* the fact,
+	 FD_WRITE will be cleared until sending is again possible.
+	 For the time being, workaround that here by using the WinSock
+	 select function.  It indicates writability correctly. */
+      if (events & FD_WRITE)
+	{
+	  wfd_set w = { 1, { fh->get_socket () } };
+	  TIMEVAL t = { 0 };
+
+	  if (_win32_select (0, NULL, &w, NULL, &t) == 0)
+	    events &= ~FD_WRITE;
+	}
+      /* Don't check for FD_CLOSE here.  Only an error case (ret == -1)
+	 will set ready for writing. */
+      me->write_ready |= ret || !!(events & (FD_WRITE | FD_CONNECT));
+    }
   if (me->except_selected)
     me->except_ready |= !!(events & FD_OOB);

^ permalink raw reply	[flat|nested] only message in thread

only message in thread, other threads:[~2023-08-01 12:23 UTC | newest]

Thread overview: (only message) (download: mbox.gz / follow: Atom feed)
-- links below jump to the message on this page --
2023-08-01 12:23 [newlib-cygwin/main] Cygwin: select: workaround FD_WRITE network event handling Corinna Vinschen

This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox;
as well as URLs for read-only IMAP folder(s) and NNTP newsgroup(s).